const Cost = require('./Cost');
const MenuAnalytics = require('./MenuAnalytics');
const Rules = require('./Rules');
const ServiceConfig = require('./ServiceConfig');

class FlexRules {
  static rules = [];
  static requiredRules = [];
  static validations = {
    PERMITE_INGREDIENTE_MESMO_DIA: Rules.ingredientDay.bind(Rules),
    PERMITE_INGREDIENTE_DIA_SEGUIDO: Rules.ingredientConsecutiveDays.bind(Rules),
    PERMITE_COR_MESMO_DIA: Rules.colorDay.bind(Rules),
    PERMITE_MODOPREPARO_MESMO_DIA: Rules.prepModeDay.bind(Rules),
    PERMITE_GRUPOPRATO_MAIOR_FREQ_MES: Rules.freqGroupMonth.bind(Rules),
    PERMITE_GRUPOPRATO_MAIOR_FREQ_SEMANA: Rules.freqGroupWeek.bind(Rules),
    PERMITE_PRATO_MAIOR_FREQ_MES: Rules.freqDishMonth.bind(Rules),
    PERMITE_PRATO_MAIOR_FREQ_SEMANA: Rules.freqDishWeek.bind(Rules),
    PERMITE_GRUPOPRATO_DIA_RESTRITO: Rules.restrictionUtilizationGroup.bind(Rules),
    PERMITE_PRATO_GRUPO_COMB_RESTRITO: Rules.restrictionDishAndGroupComb.bind(Rules),
  };

  clear() {
    FlexRules.rules = [];
    FlexRules.requiredRules = [];
  }

  setRules(rules) {
    this.clear();
    rules.forEach((rule) => {
      if (rule.IDFLEXIVEL === 'S') {
        FlexRules.rules.push(rule);
      } else {
        FlexRules.requiredRules.push(rule);
      }
    });
  }

  applyRequiredRules(dishes, otherInfo) {
    const { isFixColumn } = ServiceConfig.getConfig(otherInfo.sequential);
    if(isFixColumn) return dishes;
    return dishes.filter((dish) => {
      return FlexRules.requiredRules.every((requiredRule) => {
        return this.applyRule(requiredRule.DSREGRAFLEX, dish, otherInfo);
      });
    });
  }

  applyRule(rule, dish, otherInfo) {
    const validationMethod = FlexRules.validations[rule] || (() => {
      console.error('Invalide Rule', rule);
      return true;
    });
    return validationMethod(dish, otherInfo);
  }

  applyRulesIndividualIgnoredRules(dishes, otherInfo, rules) {
    const ignoredRules = [];
    let dishesToFilter = JSON.parse(JSON.stringify(dishes));
    let validDishes = [];
    rules.forEach(rule => {
      validDishes = dishesToFilter.filter(dish => this.applyRule(rule, dish, otherInfo));
      if(validDishes.length === 0) {
        ignoredRules.push(rule);
        validDishes = dishesToFilter;
      } else {
        dishesToFilter = validDishes;
      }
    });
    if(ignoredRules.length === rules.length) {
      validDishes = dishes;
    }
    return { dishes: validDishes, rules: ignoredRules };
  }

  applyRulesInSequency(dishes, otherInfo) {
    const rulesClone = JSON.parse(JSON.stringify(FlexRules.rules));
    const ignoredRules = [];
    const validDishes = [];
    while (ignoredRules.length < FlexRules.rules.length && validDishes.length === 0) {
      ignoredRules.push(rulesClone.shift().DSREGRAFLEX);
      dishes.forEach((dish) => {
        const isValidDish = rulesClone.every((rule) => this.applyRule(rule.DSREGRAFLEX, dish, otherInfo));
        if (isValidDish) {
          validDishes.push(dish);
        }
      });
    }
    const result = this.applyRulesIndividualIgnoredRules(validDishes, otherInfo, ignoredRules);
    return result;
  }

  analyzeMenu(menuDate, sequential, { otherDishes, todayMenu, restComb, calendar, utilizationGroupWeek }) {
    const { yesterday, tomorrow } = MenuAnalytics.getAnalyticConsecutiveDays(menuDate);
    const today = MenuAnalytics.getAnalytic(menuDate);
    const parameters = { menuDate, sequential, yesterday, today, tomorrow, todayMenu, restComb, calendar, utilizationGroupWeek };
    const filtredDishes = this.applyRequiredRules(otherDishes, parameters);
    if (!filtredDishes.length) {
      return { dish: null, rules: [] };
    }

    const { dishes, rules } = this.applyRulesInSequency(filtredDishes, parameters);
    let selectedDish = dishes[0];
    if(dishes.length > 1) {
      dishes.sort((dishA, dishB) => {
        const valueA = MenuAnalytics.data.dishCounter[dishA.prato] || 0;
        const valueB = MenuAnalytics.data.dishCounter[dishB.prato] || 0;
        if(valueA === valueB) return 0;
        return valueA < valueB ? -1 : 1
      });
      selectedDish = dishes[0];
      const minUse = MenuAnalytics.data.dishCounter[dishes[0].prato] || 0;
      for(let i = 1; i < dishes.length; i++) {
        const dishCount = (MenuAnalytics.data.dishCounter[dishes[i].prato]) || 0;
        if(dishCount === minUse) {
          if(Cost.getDishCost(dishes[i]) < Cost.getDishCost(selectedDish)) {
            selectedDish = dishes[i];
          }
        } else {
          break;
        }
      }
    }
    MenuAnalytics.analyze(menuDate, sequential, selectedDish);
    return {
      dish: selectedDish,
      rules,
    }
  }
}

module.exports = new FlexRules();
